from pyiron_base import Project
from pyiron_base import PythonTemplateJob
import os
import hashlib
import pyvista as pv
import numpy as np
from sensotwin.lifetime_extrapolation.simulation import remainingLifeTimePrediction
from sensotwin.lifetime_extrapolation.simulation import processSimulation
from ..temperature_selection.configuration import MaterialHardeningInputDataSet
from ..defect_placement.configuration import DefectInputDataSet 
from ..stress_selection.configuration import StressStrainInputDataSet 
from ..store_connection import StoreConnection

class StructureSimulation(PythonTemplateJob):
    """Pyiron job class to execute structural simulation.

        Second step in the simulation chain. 
        First calculates static loads for specified scenario and then uses those to calculate
        damage states and expected remaining lifetime for all cells/layers of the wind turbine
    """
    def __init__(self, project, job_name):
        """Initialize job."""
        super().__init__(project, job_name)
        self.input.cpu_cores=1

    def run_static(self):
        """Execute job."""
        wind_speed = self.input.stress_strain_input.get_wind_speed()
        duration = self.input.stress_strain_input.get_duration()

        # simulation expects the file path to exists before job has finished
        if not os.path.exists(self.working_directory):
            os.makedirs(self.working_directory)
        relative_job_dir = self.working_directory.replace(os.getcwd() + "/", "")
        
        mesh_list = remainingLifeTimePrediction.f_calculateRemainingLifeTime(
            wind_speed = wind_speed, 
            simulation_time = duration,
            working_dir = relative_job_dir,
            num_cpus = self.input.cpu_cores,
            ccx_input_files = self.input.ccx_input_files
        )

        for mesh in mesh_list:
            if "Fatigue" in mesh:
                self.output.mesh = mesh
            elif "Curing" in mesh:
                self.output.curing_mesh = mesh
        
        self.output.remaining_lifetime = self.extract_remaining_lifetime_from_mesh()
        self.status.finished = True
        self.to_hdf()

    def generate_ontology_label(self) -> str:
        """Generate ontology label for saving results to triple store after execution."""
        return "__LifetimeExtrapolationSimulationProcess_{material_hardening_input}_{defect_placement_input}_{stress_strain_input}".format(
            material_hardening_input = self.input.material_hardening_input.get_id(),
            defect_placement_input = self.input.defect_placement_input.get_id(),
            stress_strain_input = self.input.stress_strain_input.get_id()
        )

    def save_simulation_to_ontology(self, conn: StoreConnection):
        """Save completed job result to triple store."""
        label = self.generate_ontology_label()
        query = """
            INSERT {{
                senso:{label}_Result a senso:TotalLifeTime ;
                    co:value {remaining_lifetime} .
                senso:{label} a senso:LifetimeExtrapolationSimulationProcess ;
                    co:input senso:{material_hardening_input} ;
                    co:input senso:{defect_placement_input} ;
                    co:input senso:{stress_strain_input} ;
                    co:output senso:{label}_Result .
            }} WHERE {{ }}
            """.format(label = label, 
                       material_hardening_input = self.input.material_hardening_input.generate_ontology_label(),
                       defect_placement_input = self.input.defect_placement_input.generate_ontology_label(), 
                       stress_strain_input = self.input.stress_strain_input.generate_ontology_label(), 
                       remaining_lifetime = self.output.remaining_lifetime)
        conn.update_data(query)

    def extract_remaining_lifetime_from_mesh(self):
        """Extract maximum remaining lifetime from result mesh after job execution."""
        result_mesh = pv.read(self.output.mesh).cast_to_unstructured_grid()
        min_values = []
        for scalar in result_mesh.cell_data:
            if scalar.startswith("TFAT"):
                min_values.append(np.nanmin(result_mesh[scalar]))
        return np.min(min_values)


class MaterialSimulation(PythonTemplateJob):
    """Pyiron job class to execute material manufactoring simulation.

        First step in the simulation chain. 
        Simulates material curing behaviour and then applies selected faults to mesh input files
        for later steps
    """
    def __init__(self, project, job_name):
        """Initialize job."""
        super().__init__(project, job_name)

    def run_static(self):
        """Execute job."""
        # simulation expects the file path to exists before job has finished
        if not os.path.exists(self.working_directory):
            os.makedirs(self.working_directory)
        relative_job_dir = self.working_directory.replace(os.getcwd() + "/", "")
            
        output_files = processSimulation.SimManufac(
            fvc = self.input.material_hardening_input.get_fibre_volume_content().get_value(),
            time_data = self.input.material_hardening_input.get_temperature_graph().get_time_data(),
            temperature_data = self.input.material_hardening_input.get_temperature_graph().get_temperature_data(),
            defect_list = [x.to_simulation_input_dict() for x in self.input.defect_placement_input.get_defects()],
            working_directory = relative_job_dir
        )
        self.output.ccx_input_files = output_files
        self.status.finished = True
        self.to_hdf()


def execute_job(pr: Project,
                material_hardening_input: MaterialHardeningInputDataSet,
                defect_placement_input: DefectInputDataSet,
                stress_strain_input: StressStrainInputDataSet,
                conn: StoreConnection=None,
                cpu_cores: int=3):
    """Execute chain of pyiron jobs the calculate remaining lifetime of wind turbine according to given scenario.
    
    Args:
        pr: Initialized pyiron project object
        material_hardening_input: Parameters for material curing behaviour
        defect_placement_input: Parameters for component manufactoring defects
        stress_strain_input: Operational parameters for wind and duration of simulation
        conn: If valid StoreConnection is passed, result will be saved to specified triple store
        cpu_cores: Number of CPU cores used by CalculiX during simulation
    """
    name_hash = "{}_{}_{}".format(material_hardening_input.get_id(),defect_placement_input.get_id(), stress_strain_input.get_id())
    
    curing_job = pr.create_job(MaterialSimulation, job_name=("curing_", name_hash), delete_existing_job=True)
    curing_job.input.material_hardening_input = material_hardening_input
    curing_job.input.defect_placement_input = defect_placement_input
    curing_job.input.stress_strain_input = stress_strain_input
    curing_job.run()

    structure_job = pr.create_job(StructureSimulation, job_name=("struct_", name_hash), delete_existing_job=True)
    structure_job.input.material_hardening_input = material_hardening_input
    structure_job.input.defect_placement_input = defect_placement_input
    structure_job.input.stress_strain_input = stress_strain_input
    structure_job.input.cpu_cores = cpu_cores
    structure_job.input.ccx_input_files = curing_job.output.ccx_input_files
    structure_job.run()
    
    if conn:
        structure_job.save_simulation_to_ontology(conn)

def get_pyiron_project(project_path: str) -> Project:
    """Get pyiron project object to use for simulation jobs.

    Args:
        project_path: Relative path for pyiron project folder

    Returns:
        Initialized pyiron project object
    """
    return Project(path=project_path)

def get_job_table(pr: Project):
    return pr.job_table(auto_refresh_job_status=False)