import os
import shutil
import time
import subprocess
from datetime import datetime

from .lib import combineINP
from .lib import staticAnalysisAerodynamicsCentrifugalCCX
from .lib import staticAnalysisGravityCCX
from .lib import convertFRDtoVTK
from .lib import createElementStackingTable
from .lib import extractStresses
from .lib import fatigueAnalysisCCX

CCX_TEMPLATE_NAME = "CCX_TEMPLATE"

def f_calculateRemainingLifeTime(wind_speed=None, simulation_time=None, working_dir = './', num_cpus=1, ccx_input_files=None):
    '''
    Top level script to set up the simulations, run them and
    create the results in form of a .vtu file.

    Parameters:
    - wind_speed (float):
        Average wind speed acting on the entire rotor disc in m/s.
    - simulation_time (float):
        Total time the rotor blade is exposed to wind_speed in years.
    - num_cpus (int):
        Number of CPUs to run the CCX solver on. Defaults to 1.
    - working_dir (string):
        Working directory to execute file in.

    Returns:
    - None
    '''

    # change number of processors to run CCX with
    os.environ['OMP_NUM_THREADS'] = str(num_cpus)
        
    current_dir = working_dir
    shutil.copytree(os.path.dirname(os.path.realpath(__file__)) + "/IN", current_dir + "/IN")

    # create folder structure
    # |- IN
    # |- OUT
    # |
    # |- lib
    #    |- 02_convertFRDtoVTK_PyModule.py
    #    |- 03_createElementStackingTable.py
    #    |- 04_extractStresses.py
    #    |- readVTT.py
    # |
    # |- main
    #    |- Static
    #       |- IN
    #       |- OUT
    #    |- Fatigue
    #       |- IN
    #       |- OUT

    root_directories = ['IN','OUT','main']
    main_directories = ['Static','Fatigue']
    analysis_directories = ['IN','OUT']

    for directory in root_directories:
        if not os.path.exists(os.path.join(current_dir,directory)):
            os.mkdir(os.path.join(current_dir,directory))

    for directory in main_directories:
        if not os.path.exists(os.path.join(current_dir,'main',directory)):
            os.mkdir(os.path.join(current_dir,'main',directory))

    for analysis_directory in main_directories:
        for directory in analysis_directories:
            if not os.path.exists(os.path.join(current_dir,'main',analysis_directory,directory)):
                os.mkdir(os.path.join(current_dir,'main',analysis_directory,directory))

    # copy necessary files to folder structure
    shutil.copy(os.path.join(current_dir,'IN','aero_loads_N_transformed.vtt'),
                os.path.join(current_dir,'main','Static','IN','aero_loads_N_transformed.vtt'))
    shutil.copy(os.path.join(current_dir,'IN','gravity_loads_transformed.vtt'),
                os.path.join(current_dir,'main','Static','IN','gravity_loads_transformed.vtt'))
    shutil.copy(os.path.join(current_dir,'IN','coords_N_web_1.vtt'),
                os.path.join(current_dir,'main','Static','IN','coords_N_web_1.vtt'))
    shutil.copy(os.path.join(current_dir,'IN','coords_N_web_2.vtt'),
                os.path.join(current_dir,'main','Static','IN','coords_N_web_2.vtt'))
    shutil.copy(os.path.join(current_dir,'IN','composite_layup_db.vtt'),
                os.path.join(current_dir,'main','Static','IN','composite_layup_db.vtt'))
    shutil.copy(os.path.join(current_dir,'IN','material_fatigue_data.vtt'),
                os.path.join(current_dir,'main','Fatigue','IN','material_fatigue_data.vtt'))
    shutil.copy(os.path.join(current_dir,'IN','rpm_windspeed_0p01.vtt'),
                os.path.join(current_dir,'main','Fatigue','IN','rpm_windspeed_0p01.vtt'))
    for ccx_input_file in ccx_input_files:
        shutil.copy(os.path.join(ccx_input_file), os.path.join(current_dir,'main','Static','IN'))

    # combine input file in case *INCLUDE statements are used

    all_current_files = os.listdir(os.path.join(current_dir,'main','Static','IN'))
    all_original_inp_files = [x for x in all_current_files if x.endswith('.inp')]
    if len(all_original_inp_files) == 0:
        raise Exception("No .inp File found among input paths!")
    if len(all_original_inp_files) > 1:
        raise Exception("More than 1 .inp file found in input paths!")
    
    combineINP.f_combineINP(inp_file = all_original_inp_files[0], working_dir = os.path.join(current_dir,'main','Static','IN'), result_name = CCX_TEMPLATE_NAME)
    
    # set up static analysis
    static_dir = os.path.join(current_dir,'main','Static')

    staticAnalysisAerodynamicsCentrifugalCCX.f_setupStaticAerodynamicsCentrifugalSimulations(CCX_TEMPLATE_NAME, wind_speed, working_dir=static_dir)
    staticAnalysisGravityCCX.f_setupStaticGravitySimulations(CCX_TEMPLATE_NAME, wind_speed, working_dir=static_dir)

    # run static simulations
    all_files = os.listdir(os.path.join(static_dir,'OUT'))
    all_dirs = [x for x in all_files if os.path.isdir(os.path.join(static_dir,'OUT',x))]

    all_dirs.sort()

    for i,direc in enumerate(all_dirs):
        current_working_dir = os.path.join(static_dir,'OUT',direc)

        all_current_files = os.listdir(current_working_dir)
        all_current_inp_files = [x for x in all_current_files if x.endswith('.inp')]
        for file in all_current_inp_files:
            subprocess.run(["ccx", current_working_dir + "/" + file.strip('.inp')], capture_output=True)


            convertFRDtoVTK.f_convertFRD2VTU(current_working_dir)
            if i == 0:

                createElementStackingTable.f_createElementStackingTable(current_working_dir)
            else:
                shutil.copy(os.path.join(static_dir,'OUT',all_dirs[0],'element_layer_assignment.out.txt'),
                            os.path.join(current_working_dir,'element_layer_assignment.out.txt')) 

            extractStresses.f_extractStresses(current_working_dir)

    # run fatigue calculation
    fatigue_dir = os.path.join(current_dir,'main','Fatigue')

    fatigueAnalysisCCX.f_calculateDamageAndFatigueLife(wind_speed,simulation_time,fatigue_dir)

    result_paths = []
    fatigue_output_dir = os.path.join(current_dir,'main','Fatigue','OUT')
    all_files = os.listdir(fatigue_output_dir)
    all_result_files = [x for x in all_files if x.endswith('.vtu')]
    result_file = all_result_files[0]
    output_file = 'Fatigue_Results_Rotorblade_1_' + datetime.now().strftime('%m%d%Y_%H%M%S') + '.vtu'
    shutil.copy(os.path.join(fatigue_output_dir,result_file),os.path.join(current_dir,'OUT',output_file))
    result_paths.append(os.path.join(current_dir,'OUT',output_file))

    curing_output_dir = os.path.join(current_dir,'main','Static','OUT','AerodynamicsCentrifugal')
    all_files = os.listdir(curing_output_dir)
    all_result_files = [x for x in all_files if x.endswith('_processed_curing.vtu')]
    result_file = all_result_files[0]
    output_file = 'Curing_Results_Rotorblade_1_' + datetime.now().strftime('%m%d%Y_%H%M%S') + '.vtu'
    shutil.copy(os.path.join(curing_output_dir,result_file),os.path.join(current_dir,'OUT',output_file))
    result_paths.append(os.path.join(current_dir,'OUT',output_file))
    
    return result_paths